import re
import struct
import string
import timepro
import logging
import uuid

import sys
# Import lzma this way so we get the built in version for
# Python 3.3 or the backported one otherwize. Don't just
# do a try/catch for import lzma because the older
# pyliblzma uses that package name, and we do not want
# to use it.
if sys.version_info[0:3] >= (3,3,0):
    import lzma
else:
    try:
        from backports import lzma
    except ImportError as e:
        # On Fedora/OLPC we have a namespace conflict
        # with another package, so we had to rename the
        # namespace until we get it resolved upstream
        from backportslzma import lzma

try:
    # repose.lru does not help much in production
    # web serving.  It makes a HUGE difference in
    # pre-processing.  So we don't want to make it
    # a hard dependency.
    from repoze.lru import LRUCache
except ImportError as e:
    class LRUCache(object):
        def __init__(self, cache_size):
            return

        def get(self, key, value=None):
            return value

        def put(self, key, value):
            return

        def clear(self):
            return

from StringIO import StringIO

logger = logging.getLogger(__name__)

HEADER_FORMAT = [
    ('I', 'magicNumber'),
    ('I', 'version'),
    #('Q', 'uuidLow'),
    #('Q', 'uuidHigh'),
    ('B', 'uuid0'),
    ('B', 'uuid1'),
    ('B', 'uuid2'),
    ('B', 'uuid3'),
    ('B', 'uuid4'),
    ('B', 'uuid5'),
    ('B', 'uuid6'),
    ('B', 'uuid7'),
    ('B', 'uuid8'),
    ('B', 'uuid9'),
    ('B', 'uuid10'),
    ('B', 'uuid11'),
    ('B', 'uuid12'),
    ('B', 'uuid13'),
    ('B', 'uuid14'),
    ('B', 'uuid15'),
    ('I', 'articleCount'),
    ('I', 'clusterCount'),
    ('Q', 'urlPtrPos'),
    ('Q', 'titlePtrPos'),
    ('Q', 'clusterPtrPos'),
    ('Q', 'mimeListPos'),
    ('I', 'mainPage'),
    ('I', 'layoutPage'),
    ('Q', 'checksumPos')
]


ARTICLE_ENTRY_FORMAT = [
    ('H', 'mimetype'),
    ('B', 'parameterLen'),
    ('c', 'namespace'),
    ('I', 'revision'),
    ('I', 'clusterNumber'),
    ('I', 'blobNumber')
    # Null terminated url
    # Null terminated title
    # variable length parameter data
]


REDIRECT_ENTRY_FORMAT = [
    ('H', 'mimetype'),
    ('B', 'parameterLen'),
    ('c', 'namespace'),
    ('I', 'revision'),
    ('I', 'redirectIndex')
    # Null terminated url
    # Null terminated title
    # variable length parameter data
]


CLUSTER_FORMAT = [
    ('B', 'compressionType')
]


# A null byte
NULL = struct.pack('B', 0)


def format_from_rich(rich_format):
    return "<" + string.join([x[0] for x in rich_format], "")


@timepro.profile()
def read_null_terminated(f, encoding='utf-8'):
    s = ""
    while True:
        b = f.read(1)
        if b == NULL:
            return s.decode(encoding)
        s = s + b


def binary_search(f, t, min, max):
    while 1:
        if max < min:
            return None
        m = (min + max) / 2
        v = f(m)
        if v < t:
            min = m + 1
        elif v > t:
            max = m - 1
        else:
            return m


def full_url(namespace, url):
    return namespace + '/' + url


class Format(object):
    def __init__(self, rich_format):
        self.rich_fmt = rich_format
        self.fmt = format_from_rich(rich_format)
        self.compiled = struct.Struct(self.fmt)
        self.size = self.compiled.size

    @timepro.profile()
    def unpack_format(self, buffer, offset=0):
        fields = self.compiled.unpack_from(buffer, offset)
        d = []
        for field, entry in zip(fields, self.rich_fmt):
            d.append((entry[1], field))
        return d

    @timepro.profile()
    def unpack_format_from_file(self, f, seek=None):
        if seek is not None:
            timepro.start("seek")
            f.seek(seek)
            timepro.end("seek")
        buf = f.read(self.size)
        d = self.unpack_format(buf)
        return d

    @timepro.profile()
    def unpack(self, buffer, offset=0):
        """Override this to get more complex behavior"""
        return self.unpack_format(buffer, offset)

    @timepro.profile()
    def unpack_from_file(self, f, seek=None):
        """Override this to get more complex behavior"""
        return self.unpack_format_from_file(f, seek)


class HeaderFormat(Format):
    def __init__(self):
        super(HeaderFormat, self).__init__(HEADER_FORMAT)


class ClusterFormat(Format):
    def __init__(self):
        super(ClusterFormat, self).__init__(CLUSTER_FORMAT)


class ClusterCache(object):
    def __init__(self, cache_size):
        self.lru = LRUCache(cache_size)
        self.hits = 0
        self.misses = 0

    def get(self, file_buffer, ptr):
        v = self.lru.get((file_buffer, ptr))
        if v is not None:
            self.hits += 1
            return v
        v = ClusterData(file_buffer, ptr)
        self.lru.put((file_buffer, ptr), v)
        self.misses += 1
        return v

    def clear(self):
        print "CACHE HITS " + str(self.hits) + " VS MISSES " + str(self.misses)
        self.lru.clear()


class ClusterData(object):
    @timepro.profile()
    def __init__(self, file_buffer, ptr):
        cluster_info = dict(ClusterFormat().unpack_from_file(file_buffer, ptr))
        self.compressed = cluster_info['compressionType'] == 4

        self.file_buf = file_buffer
        self.uncomp_buf = None
        self.ptr = ptr

        self.offsets = []

        if self.compressed:
            self._decompress()

        self.read_offsets()

    @timepro.profile()
    def _decompress(self, chunk_size=32000):
        """Decompresses the cluster if compression flag was found. Stores
        uncompressed results internally."""

        if not self.compressed:
            return

        self.file_buf.seek(self.ptr + 1)

        # Store uncompressed cluster data for use as uncompressed data
        self.uncomp_buf = StringIO()

        decomp = lzma.LZMADecompressor()
        while not decomp.eof:
            timepro.start("file_buf read")
            comp_data = self.file_buf.read(chunk_size)
            timepro.end("file_buf read")

            timepro.start("decompress")
            uncomp_data = decomp.decompress(comp_data)
            timepro.end("decompress")

            timepro.start("write")
            self.uncomp_buf.write(uncomp_data)
            timepro.end("write")

        return self.uncomp_buf

    def source_buffer(self):
        """Returns the buffer to read from, either the file buffer
        passed or the uncompressed lzma data. Will seek to the
        beginning of the cluster after the 1 byte compression flag"""

        if self.compressed:
            self.uncomp_buf.seek(0)
            return self.uncomp_buf
        else:
            self.file_buf.seek(self.ptr + 1)
            return self.file_buf

    def unpack_blob_index(self, buf):
        ptr = struct.unpack('I', buf)[0]
        return ptr

    @timepro.profile()
    def read_offsets(self):
        """Reads the cluster header with the offsets of the blobs"""

        src_buf = self.source_buffer()

        raw = src_buf.read(4)
        offset0 = self.unpack_blob_index(raw)
        self.offsets.append(offset0)
        nblob = offset0 / 4

        for idx in range(nblob-1):
            raw = src_buf.read(4)
            offset = self.unpack_blob_index(raw)
            self.offsets.append(offset)

        return self.offsets

    @timepro.profile()
    def read_blob(self, blob_index):
        """Reads a blob from the cluster"""

        if blob_index >= len(self.offsets) - 1:
            raise IOError("Blob index exceeds number of blobs available: %s" % blob_index)

        src_buf = self.source_buffer()

        blob_size = self.offsets[blob_index+1] - self.offsets[blob_index]

        # For uncompressed data, seek from beginning of file
        # Otherwise seek the compressed data with just and offset
        if not self.compressed:
            seek_beg = self.ptr + 1
        else:
            seek_beg = 0
        src_buf.seek(seek_beg + self.offsets[blob_index])

        blob_data = src_buf.read(blob_size)

        return blob_data

class ArticleEntryFormat(Format):
    def __init__(self):
        super(ArticleEntryFormat, self).__init__(ARTICLE_ENTRY_FORMAT)

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        d = super(ArticleEntryFormat, self).unpack_from_file(f, seek)
        url = read_null_terminated(f)
        title = read_null_terminated(f)
        parameter = f.read(dict(d)['parameterLen'])
        d.extend([('url', url),
                  ('title', title),
                  ('parameter', parameter)]
                 )
        return d


class RedirectEntryFormat(Format):
    def __init__(self):
        super(RedirectEntryFormat, self).__init__(REDIRECT_ENTRY_FORMAT)

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        d = super(RedirectEntryFormat, self).unpack_from_file(f, seek)
        url = read_null_terminated(f)
        title = read_null_terminated(f)
        parameter = f.read(dict(d)['parameterLen'])
        d.extend([('url', url),
                  ('title', title),
                  ('parameter', parameter)]
                 )
        return d


class MimeTypeListFormat(Format):
    def __init__(self):
        super(MimeTypeListFormat, self).__init__("")

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        if seek is not None:
            f.seek(seek)
        mimetypes = []
        while True:
            s = read_null_terminated(f)
            if s == "":
                return mimetypes
            mimetypes.append(s)


class ZimFile(object):
    def __init__(self, filename, cache_size=4):
        self.filename = filename
        self.redirectEntryFormat = RedirectEntryFormat()
        self.articleEntryFormat = ArticleEntryFormat()
        self.clusterFormat = ClusterFormat()
        self.f = open(filename, "r")
        self.header = dict(HeaderFormat().unpack_from_file(self.f))
        self.mimeTypeList = MimeTypeListFormat().unpack_from_file(self.f, self.header['mimeListPos'])
        self.clusterCache = ClusterCache(cache_size=cache_size)

    def close(self):
        self.clusterCache.clear()
        self.f.close()

    def get_uuid(self):
        """Returns the UUID for this ZIM file"""
        h = self.header
        uuid_bytes = [h['uuid0'], h['uuid1'], h['uuid2'], h['uuid3'], h['uuid4'],
                      h['uuid5'], h['uuid6'], h['uuid7'], h['uuid8'], h['uuid9'],
                      h['uuid10'], h['uuid11'], h['uuid12'], h['uuid13'], h['uuid14'],
                      h['uuid15']]
        s = string.join([chr(x) for x in uuid_bytes], "")
        return uuid.UUID(bytes=s)

    def get_kiwix_uuid(self):
        """Kiwix seems to have a bug in their library.xml which causes the
        third UUID group to be repeated."""
        u = self.get_uuid()
        s = str(u).split("-")
        return s[0] + "-" + s[1] + "-" + s[2] + "-" + s[2] + "-" + s[3] + s[4]

    @timepro.profile()
    def read_directory_entry(self, offset):
        """May return either a Redirect or Article entry depending on flag"""
        timepro.start("seek")
        self.f.seek(offset)
        timepro.end("seek")
        timepro.start("read(2)")
        buf = self.f.read(2)
        timepro.end("read(2)")
        fields = struct.unpack('H', buf)
        if fields[0] == 0xffff:  # Then redirect
            return dict(self.redirectEntryFormat.unpack_from_file(self.f, offset))
        else:
            return dict(self.articleEntryFormat.unpack_from_file(self.f, offset))

    @timepro.profile()
    def read_url_pointer(self, index):
        self.f.seek(self.header['urlPtrPos'] + 8 * index)
        buf = self.f.read(8)
        fields = struct.unpack('Q', buf)
        return fields[0]

    def read_title_pointer(self, index):
        self.f.seek(self.header['titlePtrPos'] + 4 * index)
        buf = self.f.read(4)
        fields = struct.unpack('L', buf)
        return fields[0]

    @timepro.profile()
    def read_cluster_pointer(self, index):
        """Returns a pointer to the cluster"""

        self.f.seek(self.header['clusterPtrPos'] + 8 * index)
        buf = self.f.read(8)
        fields = struct.unpack('Q', buf)
        return fields[0]

    @timepro.profile()
    def read_directory_entry_by_index(self, index):
        ptr = self.read_url_pointer(index)
        d = self.read_directory_entry(ptr)
        d['index'] = index
        return d

    @timepro.profile()
    def read_blob(self, cluster_index, blob_index):
        ptr = self.read_cluster_pointer(cluster_index)
        cluster_data = self.clusterCache.get(self.f, ptr)
        return cluster_data.read_blob(blob_index)

    @timepro.profile()
    def get_article_by_index(self, index, follow_redirect=True):
        entry = self.read_directory_entry_by_index(index)
        if 'redirectIndex' in entry.keys():
            if follow_redirect:
                logger.debug("REDIRECT TO " + str(entry['redirectIndex']))
                return self.get_article_by_index(entry['redirectIndex'], follow_redirect)
            else:
                return None, entry['redirectIndex'], entry['namespace']
        data = self.read_blob(entry['clusterNumber'], entry['blobNumber'])
        mime = self.mimeTypeList[entry['mimetype']]
        namespace = entry['namespace']
        return data, mime, namespace

    @timepro.profile()
    def get_entry_by_url_linear(self, namespace, url):
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            if entry['url'] == url and entry['namespace'] == namespace:
                return i
        return None

    @timepro.profile()
    def get_entry_by_url(self, namespace, url):
        nsurl = full_url(namespace, url)

        def check(idx):
            entry = self.read_directory_entry_by_index(idx)
            return full_url(entry['namespace'], entry['url'])

        m = binary_search(check, nsurl, 0, self.header['articleCount'])
        if m is None:
            return None, None
        entry = self.read_directory_entry_by_index(m)
        return entry, m

    def get_article_by_url(self, namespace, url, follow_redirect=True):
        entry, idx = self.get_entry_by_url(namespace, url)
        if idx is None:
            return None, None, None
        return self.get_article_by_index(idx, follow_redirect=follow_redirect)

    def get_main_page(self):
        main_index = self.header['mainPage']
        return self.get_article_by_index(main_index)

    @timepro.profile()
    def metadata(self):
        metadata = {}
        for i in xrange(self.header['articleCount'] - 1, -1, -1):
            entry = self.read_directory_entry_by_index(i)
            if entry['namespace'] == 'M':
                m_name = entry['url']
                # Lower case first letter to match kiwix-library names convention
                m_name = re.sub(r'^([A-Z])', lambda pat: pat.group(1).lower(), m_name)
                metadata[m_name] = self.get_article_by_index(i)[0]
            else:
                break

        return metadata

    def articles(self):
        """Generator which iterates through all articles"""
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            entry['fullUrl'] = full_url(entry['namespace'], entry['url'])
            yield entry

    def validate(self):
        """This is a mostly a self-test, but will validate various assumptions"""
        # Test that URLs are properly ordered
        last = None
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            assert entry is not None
            nsurl = full_url(entry['namespace'], entry['url'])
            if last is not None:
                assert nsurl > last
            last = nsurl
        timepro.log_all()
        timepro.reset()

        # Test load by url performance
        for i in range(0, self.header['articleCount'], 100):
            entry = self.read_directory_entry_by_index(i)
            entry2, idx = self.get_entry_by_url(entry['namespace'], entry['url'])
            assert entry2 is not None
        timepro.log_all()
        timepro.reset()

        # Test load of the last article
        article, mime, ns = self.get_article_by_index(self.header['articleCount'] - 1)
        entry = self.read_directory_entry_by_index(self.header['articleCount'] - 1)
        entry2, idx = self.get_entry_by_url(entry['namespace'], entry['url'])
        assert entry2 is not None

        # Test load subset of all articles
        for i in range(0, self.header['articleCount'], 100):
            if i % 1000 == 0:
                print i
            article, mime, ns = self.get_article_by_index(i)
            if article is None:  # Redirect
                assert mime is not None
        timepro.log_all()
        timepro.reset()

    def list_articles_by_url(self):
        """Mostly for testing"""
        s = ""
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            s += full_url(entry['namespace'], entry['url']) + "\n"
        return s



#autogenerated by sqlautocode

from .extensions import db

gutenberg_books_category_map = db.Table(u'gutenberg_books_category_map', 
    db.Column(u'book_id', db.INTEGER(), db.ForeignKey('gutenberg_books.textId'), primary_key=True),
    db.Column(u'category_id', db.INTEGER(), db.ForeignKey('gutenberg_categories.id'), primary_key=True),
)

gutenberg_books_contributor_map = db.Table(u'gutenberg_books_contributor_map', 
    db.Column(u'book_id', db.INTEGER(), db.ForeignKey('gutenberg_books.textId'), primary_key=True),
    db.Column(u'contributor_id', db.INTEGER(), db.ForeignKey('gutenberg_contributors.id'), primary_key=True),
)

gutenberg_books_creator_map = db.Table(u'gutenberg_books_creator_map', 
    db.Column(u'book_id', db.INTEGER(), db.ForeignKey('gutenberg_books.textId'), primary_key=True),
    db.Column(u'creator_id', db.INTEGER(), db.ForeignKey('gutenberg_creators.id'), primary_key=True),
)

gutenberg_books_language_map = db.Table(u'gutenberg_books_language_map', 
    db.Column(u'book_id', db.INTEGER(), db.ForeignKey('gutenberg_books.textId'), primary_key=True),
    db.Column(u'language_id', db.INTEGER(), db.ForeignKey('gutenberg_languages.id'), primary_key=True),
)

gutenberg_books_subject_map = db.Table(u'gutenberg_books_subject_map', 
    db.Column(u'book_id', db.INTEGER(), db.ForeignKey('gutenberg_books.textId'), primary_key=True),
    db.Column(u'subject_id', db.INTEGER(), db.ForeignKey('gutenberg_subjects.id'), primary_key=True),
)

class GutenbergBook(db.Model):
    __tablename__ = 'gutenberg_books'

    __table_args__ = {}

    #column definitions
    downloads = db.Column(u'downloads', db.INTEGER())
    friendlytitle = db.Column(u'friendlytitle', db.TEXT())
    textId = db.Column(u'textId', db.TEXT(), primary_key=True, nullable=False)
    title = db.Column(u'title', db.TEXT())
    title_order = db.Column(u'title_order', db.INTEGER())

    #relation definitions
    gutenberg_categories = db.relation('GutenbergCategory', primaryjoin='GutenbergBook.textId==gutenberg_books_category_map.c.book_id', secondary=gutenberg_books_category_map, secondaryjoin='gutenberg_books_category_map.c.category_id==GutenbergCategory.id')
    gutenberg_contributors = db.relation('GutenbergContributor', primaryjoin='GutenbergBook.textId==gutenberg_books_contributor_map.c.book_id', secondary=gutenberg_books_contributor_map, secondaryjoin='gutenberg_books_contributor_map.c.contributor_id==GutenbergContributor.id')
    gutenberg_creators = db.relation('GutenbergCreator', primaryjoin='GutenbergBook.textId==gutenberg_books_creator_map.c.book_id', secondary=gutenberg_books_creator_map, secondaryjoin='gutenberg_books_creator_map.c.creator_id==GutenbergCreator.id')
    gutenberg_languages = db.relation('GutenbergLanguage', primaryjoin='GutenbergBook.textId==gutenberg_books_language_map.c.book_id', secondary=gutenberg_books_language_map, secondaryjoin='gutenberg_books_language_map.c.language_id==GutenbergLanguage.id')
    gutenberg_subjects = db.relation('GutenbergSubject', primaryjoin='GutenbergBook.textId==gutenberg_books_subject_map.c.book_id', secondary=gutenberg_books_subject_map, secondaryjoin='gutenberg_books_subject_map.c.subject_id==GutenbergSubject.id')
    gutenberg_files = db.relation('GutenbergFile')


class GutenbergCategory(db.Model):
    __tablename__ = 'gutenberg_categories'

    __table_args__ = {}

    #column definitions
    category = db.Column(u'category', db.TEXT())
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergCategory.id==gutenberg_books_category_map.c.category_id', secondary=gutenberg_books_category_map, secondaryjoin='gutenberg_books_category_map.c.book_id==GutenbergBook.textId')


class GutenbergContributor(db.Model):
    __tablename__ = 'gutenberg_contributors'

    __table_args__ = {}

    #column definitions
    contributor = db.Column(u'contributor', db.TEXT())
    downloads = db.Column(u'downloads', db.INTEGER())
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergContributor.id==gutenberg_books_contributor_map.c.contributor_id', secondary=gutenberg_books_contributor_map, secondaryjoin='gutenberg_books_contributor_map.c.book_id==GutenbergBook.textId')


class GutenbergCreator(db.Model):
    __tablename__ = 'gutenberg_creators'

    __table_args__ = {}

    #column definitions
    creator = db.Column(u'creator', db.TEXT())
    downloads = db.Column(u'downloads', db.INTEGER())
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergCreator.id==gutenberg_books_creator_map.c.creator_id', secondary=gutenberg_books_creator_map, secondaryjoin='gutenberg_books_creator_map.c.book_id==GutenbergBook.textId')


class GutenbergFile(db.Model):
    __tablename__ = 'gutenberg_files'

    __table_args__ = {}

    #column definitions
    file = db.Column(u'file', db.TEXT())
    format = db.Column(u'format', db.TEXT())
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)
    textId = db.Column(u'textId', db.TEXT(), db.ForeignKey('gutenberg_books.textId'))

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergFile.textId==GutenbergBook.textId')


class GutenbergLanguage(db.Model):
    __tablename__ = 'gutenberg_languages'

    __table_args__ = {}

    #column definitions
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)
    language = db.Column(u'language', db.TEXT())

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergLanguage.id==gutenberg_books_language_map.c.language_id', secondary=gutenberg_books_language_map, secondaryjoin='gutenberg_books_language_map.c.book_id==GutenbergBook.textId')


class GutenbergSubject(db.Model):
    __tablename__ = 'gutenberg_subjects'

    __table_args__ = {}

    #column definitions
    id = db.Column(u'id', db.INTEGER(), primary_key=True, nullable=False)
    subject = db.Column(u'subject', db.TEXT())

    #relation definitions
    gutenberg_books = db.relation('GutenbergBook', primaryjoin='GutenbergSubject.id==gutenberg_books_subject_map.c.subject_id', secondary=gutenberg_books_subject_map, secondaryjoin='gutenberg_books_subject_map.c.book_id==GutenbergBook.textId')





# Utility functions for Gutenberg content
# such as htmlz and epub datasets
import os

from config import config


def hashdir(n):
    return "%02i" % (int(n) % 100)


def build_htmlz_filename(n):
    return "pg%i.htmlz" % n


def build_epub_filename(n):
    return "pg%i.epub" % n


def find_htmlz(pgid):
    """Find an htmlz file from a gutenberg book id.
    Prefers htmlz with images if available, otherwise
    falls back to non-image version.  Returns None if
    no content exists"""
    htmlz_dir = config().get_path('GUTENBERG', 'htmlz_dir')
    htmlz_images_dir = config().get_path('GUTENBERG', 'htmlz_images_dir')
    hashpath = hashdir(pgid)
    filename = build_htmlz_filename(pgid)
    hashpath = os.path.join(hashpath, filename)
    htmlz_path = os.path.join(htmlz_images_dir, hashpath)
    if not os.path.exists(htmlz_path):
        htmlz_path = os.path.join(htmlz_dir, hashpath)
    if not os.path.exists(htmlz_path):
        return None
    return htmlz_path


def find_epub(pgid):
    """Find an epub file from a gutenberg book id.
    Prefers epub with images if available, otherwise
    falls back to non-image version.  Returns None if
    no content exists"""
    epub_dir = config().get_path('GUTENBERG', 'epub_dir')
    epub_images_dir = config().get_path('GUTENBERG', 'epub_images_dir')
    hashpath = hashdir(pgid)
    filename = build_epub_filename(pgid)
    hashpath = os.path.join(hashpath, filename)
    epub_path = os.path.join(epub_images_dir, hashpath)
    if not os.path.exists(epub_path):
        epub_path = os.path.join(epub_dir, hashpath)
    if not os.path.exists(epub_path):
        return None
    return epub_path



# Search views
from flask import Blueprint, Response, request, redirect, make_response
import json

from wikipedia_search import WikipediaSearch
from map_search import MapSearch
from config import config

blueprint = Blueprint('search_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route("search_counts", methods=['GET'])
def search_counts_view():
    """Returns JSON containing the counts of matches
    in each major type of search"""
    query = request.args.get('q')
    counts = {}

    # Query wikipedia titles for matches
    ws = WikipediaSearch("wikititles_index")
    counts['wikipedia'] = ws.count(query)
    # Add additional search types here

    # Dump all matches to JSON
    j = json.dumps(counts, indent=4)
    return Response(j, mimetype='application/json')


@blueprint.route('search_wikipedia', methods=['GET'])
def search_wikipedia_view():
    """Return JSON containing search results for
    Wikipedia index"""
    query = request.args.get('q')
    pagelen = request.args.get('pagelen', 0, int)
    page = request.args.get('page', 1, int)
    ws = WikipediaSearch("wikititles_index")
    results = ws.search(query, pagelen=pagelen, page=page)
    j = json.dumps(results, indent=4)
    return Response(j, mimetype='application/json')

@blueprint.route('search_maps', methods=['GET'])
def search_map_view():
    query = request.args.get('q')
    pagelen = request.args.get('pagelen', 0, int)
    page = request.args.get('page', 1, int)
    path = config().get_path('OSM', 'osm_search_dir')
    ms = MapSearch(path)
    results = ms.search(query, pagelen=pagelen, page=page)
    j = json.dumps(results, indent=4)
    return Response(j, mimetype='application/json')



from flask import Flask, url_for
from flask.ext.babel import Babel
from flask.ext.autoindex import AutoIndex
import sys
import urlparse

from config import config
import top_views
import search_views
import map_views
import video_views
import gutenberg
import gutenberg_content_views
import wikipedia_views
import zim_views
import settings_views
from babel_patch import babel_patched_load


def create_app(debug=True, enable_profiler=False, profiler_quiet=False, enable_timepro=False):
    """
    :param debug: whether to configure app for debug
    :param enable_profiler: enable flask profiler, causes function to return ProfilerMiddleware rather than Flask object which can be started by run_simple
    :param profiler_quiet: when profiler enabled sets whether profiler output echoed to stdout
    """
    app = Flask("iiab")
    app.url_map.strict_slashes = False
    app.use_x_sendfile = config().getboolean('WEBAPP', 'use_x_sendfile')
    app.secret_key = '1785132b4fd244a2a1ce6ae3f1d978ac'

    # Configuration items
    if debug:
        app.debug = True
        app.config['DEBUG'] = True
        app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0

    base_prefix = config().get('WEBAPP', 'base_prefix')

    blueprints = [
        (top_views.blueprint, base_prefix),
        (search_views.blueprint, base_prefix),
        (gutenberg.gutenberg, base_prefix + "books"),
        (map_views.blueprint, base_prefix + "maps"),
        (video_views.blueprint, base_prefix + "video"),
        (wikipedia_views.blueprint, base_prefix + "wikipedia"),
        (zim_views.blueprint, base_prefix + "zim"),
        (gutenberg_content_views.blueprint, base_prefix + "books"),
        (settings_views.blueprint, base_prefix + "settings"),
    ]
    for blueprint, prefix in blueprints:
        app.register_blueprint(blueprint, url_prefix=prefix)

    gutenberg.set_flask_app(app)
    gutenberg.init_db()

    configure_babel(app)

    # Auto Index the software repository
    autoindex = AutoIndex(app, add_url_rules=False)
    # FIXME: this should be done more elegantly -bcg'13

    @app.route(base_prefix + 'software/<path:path>')
    @app.route(base_prefix + 'software')
    def software_view(path='.'):
        software_dir = config().get_path('SOFTWARE', 'software_dir')
        return autoindex.render_autoindex(path, browse_root=software_dir, endpoint='software_view')

    #print "URL MAP: ", app.url_map

    # Static handling from http://flask.pocoo.org/mailinglist/archive/2011/8/25/static-files-subdomains/#9237e5b3c217b2875c59daaac4c23487
    app.config['STATIC_ROOT'] = config().get_default('WEBAPP', 'static_url_path', None)

    def static(path):
        root = app.config.get('STATIC_ROOT', None)
        if root is None:  # fallback on the normal way
            return url_for('static', filename=path)
        return urlparse.urljoin(root, path)

    @app.context_processor
    def inject_static():
        return dict(static=static)

    if enable_profiler:
        from werkzeug.contrib.profiler import ProfilerMiddleware, MergeStream
        f = open('profiler.log', 'w')
        if profiler_quiet:
            profile_app = ProfilerMiddleware(app, f, profile_dir="/tmp")
        else:
            stream = MergeStream(sys.stdout, f)
            profile_app = ProfilerMiddleware(app, stream, profile_dir="/tmp")
        return profile_app

    if enable_timepro:
        from timepro_flask import TimeProMiddleware, MergeStream
        f = open('timepro.log', 'w')
        if profiler_quiet:
            profile_app = TimeProMiddleware(app, f, profile_dir="/tmp")
        else:
            stream = MergeStream(sys.stdout, f)
            profile_app = TimeProMiddleware(app, stream, profile_dir="/tmp")
        return profile_app


    return app


def configure_babel(app):
    # Load additional languages into the babel
    # cache.  We do this so we can add Haitian Creole
    # using french as a model.
    babel_patched_load('cpf')

    # flask-babel
    babel = Babel(app)

    # Wire Babel into the settings
    # and the settings code into Babel
    settings_views.set_babel(babel)
    babel.localeselector(settings_views.current_locale)

    return babel



import os

from utils import whoosh_open_dir_32_or_64
from whoosh.qparser import MultifieldParser
from whoosh import scoring

from .whoosh_multi_field_spelling_correction import MultiFieldQueryCorrector
import pagination_helper

def index_directory_path(base_path, zim_name):
    """Returns the directory where a ZIM file's index should be located, given
    a base path where all the index files are located as well as a filename
    or partial filename of the zim file.
    """

    index_dir = os.path.join(base_path, os.path.splitext(os.path.basename(zim_name))[0])
    return index_dir


def get_query_corrections(searcher, query, qstring):
    """
    Suggest alternate spelling for search terms by searching each column with
    spelling correction support in turn.

    :param searcher: whoosh searcher object
    :param query: whoosh query object
    :param qstring: search string that was passed to the query object
    :returns: MultiFieldQueryCorrector with one corrector for each corrected column
    """
    fieldnames = [name for name, field in searcher.schema.items() if field.spelling]
    correctors = {}
    for fieldname in fieldnames:
        if fieldname not in correctors:
            correctors[fieldname] = searcher.corrector(fieldname)
    terms = []
    for token in query.all_tokens():
        if token.fieldname in correctors:
            terms.append((token.fieldname, token.text))

    return MultiFieldQueryCorrector(correctors, terms, prefix=2, maxdist=1).correct_query(query, qstring)

def deduplicate_corrections(corrections):
    """
    Return list of correction that omits entries where the query is unmodified
    :param corrections: list of Corrector objects
    :returns: list of Corrector objects
    """
    # Using values from a dictionary comprehension rather than a list comprehension in order to deduplicate
    #return {c.string : c for c in corrections if c.original_query != c.query}.values()
    # We can't use dictionary comprehension because we are stuck on python 2.6 for Debian stable
    return dict((c.string, c) for c in corrections if c.original_query != c.query).values()


def paginated_search(index_dir, search_columns, query_text, page=1, pagelen=20, sort_column=None, weighting=scoring.BM25F):
    """
    Return a tuple consisting of an object that emulates an SQLAlchemy pagination object and corrected query suggestion
    pagelen specifies number of hits per page
    page specifies page of results (first page is 1)
    """
    query_text = unicode(query_text)  # Must be unicode
    ix = whoosh_open_dir_32_or_64(index_dir)
    with ix.searcher(weighting=weighting) as searcher:
        query = MultifieldParser(search_columns, ix.schema).parse(query_text)
        try:
            # search_page returns whoosh.searching.ResultsPage
            results = searcher.search_page(query, page, pagelen=pagelen, sortedby=sort_column)
            total = results.total
        except ValueError:  # Invalid page number
            results = []
            total = 0
        paginate = pagination_helper.Pagination(page, pagelen, total, [dict(r.items()) for r in results])
        corrections = deduplicate_corrections(get_query_corrections(searcher, query, query_text))  # list of Corrector objects

        #hf = whoosh.highlight.HtmlFormatter(classname="change")
        #html = corrections.format_string(hf)
        return (paginate, [c.string for c in corrections])



from setuptools import setup
import iiab

setup(
    name='Internet-in-a-Box',
    version=iiab.__version__,
    author="Braddock Gaskill",
    author_email="braddock@braddock.com",
    license="2-clause BSD license",
    description="Sharing the world's Free information",
    long_description=open('README.md').read(),
    url="http://internet-in-a-box.org",
    packages=['iiab'],
    scripts=['iiab-server', 'iiab.wsgi'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Flask >= 0.10',
        'Flask-Babel >= 0.8',
        'Flask-AutoIndex',
        'Flask-SQLAlchemy',
        'SQLAlchemy >= 0.8.2',
        'whoosh >= 2.4.1',
        'backports.lzma >= 0.0.2',
        'repoze.lru'
    ]
    #data_files=[("", ["LICENSE.txt", "README.md", "INSTALL.txt"])]
)



# -*- coding: utf-8 -*-

import os
import re

from flask import (Blueprint, render_template, request, Response,
                   flash, url_for, redirect, safe_join, make_response,
                   send_file, send_from_directory)
from flask.ext.babel import gettext as _
import json

from contextlib import closing

from .extensions import db
from gutenberg_models import (GutenbergBook, GutenbergFile,
                              GutenbergCreator, gutenberg_books_creator_map)
from gutenberg_content import find_htmlz, find_epub
from config import config

from whoosh_search import paginated_search
from whoosh.index import open_dir
from .endpoint_description import EndPointDescription

DEFAULT_RESULTS_PER_PAGE = 20
DEFAULT_SEARCH_COLUMNS = ['title', 'creator', 'contributor']  # names correspond to fields in whoosh schema

gutenberg = Blueprint('gutenberg', __name__, url_prefix='/books')
etext_regex = re.compile(r'^etext(\d+)$')


flask_app = None
is_init = False


def set_flask_app(app):
    global flask_app
    flask_app = app


def init_db():
    global is_init
    global flask_app
    if not is_init:
        if flask_app is None:
            raise Exception("init_db called when flask app not set")
        # set global config variables referenced by SQLAlchemy
        flask_app.config['SQLALCHEMY_ECHO'] = config().getboolean('GUTENBERG', 'sqlalchemy_echo')
        database_path = config().get_path('GUTENBERG', 'sqlalchemy_database_uri')
        flask_app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///' + os.path.abspath(database_path)
        db.init_app(flask_app)
        is_init = True


@gutenberg.route('/')
def index():
    return render_template('gutenberg/index.html')


@gutenberg.route('/search')
def search():
    query = request.args.get('q', '').strip()
    pagination = None
    if query:
        index_dir = config().get_path('GUTENBERG', 'index_dir')
        page = int(request.args.get('page', 1))
        (pagination, suggestion) = paginated_search(index_dir, DEFAULT_SEARCH_COLUMNS, query, page, sort_column='creator')
    else:
        flash(_('Please input keyword(s)'), 'error')
    #print pagination.items
    return render_template('gutenberg/search.html', pagination=pagination,
                           keywords=query, suggestion=suggestion, fn_author_to_query=author_to_query,
                           endpoint_desc=EndPointDescription('gutenberg.search', None),
                           files_exist=files_exist)


@gutenberg.route('/mirror/<path:filename>')
def gutenberg_mirror(filename):
    mirror_dir = config().get_path('GUTENBERG', 'gutenberg_mirror')
    print "mirror", mirror_dir, filename
    r = send_from_directory(mirror_dir, filename)
    return make_response(r, 200, {'Accept-Ranges': 'bytes'})


def author_to_query(author):
    """Helper function for template macro to convert an author string into a
    search query.

    :param author: unicode string from creator or contributor field of gutenberg index
    :returns: whoosh query string to search creator/contributor fields for given author.
    """
    # contributor field provides extra details about contirbutor's role in brackets -- strip that off so we can search for author in any role.
    author = re.sub(r'\[[^\]]+\]', '', author).strip()
    return u'creator:"{0}" OR contributor:"{0}"'.format(author)


@gutenberg.route('/titles')
def by_title():
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergBook.query.order_by(GutenbergBook.title_order).paginate(page, per_page)
    return render_template('gutenberg/title-index.html', pagination=pagination, fn_author_to_query=author_to_query, endpoint_desc=EndPointDescription('.by_title', dict(per_page=per_page)))


@gutenberg.route('/authors')
def by_author():
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergCreator.query.order_by(GutenbergCreator.creator).paginate(page, per_page)

    return render_template('gutenberg/author-index.html', pagination=pagination, endpoint_desc=EndPointDescription('.by_author', dict(per_page=per_page)))


@gutenberg.route('/author/<authorId>')
def author(authorId):
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergBook.query.filter(gutenberg_books_creator_map.c.creator_id == authorId).filter(gutenberg_books_creator_map.c.book_id == GutenbergBook.textId).paginate(page, per_page)
    return render_template('gutenberg/title-index.html', pagination=pagination, fn_author_to_query=author_to_query, endpoint_desc=EndPointDescription('.author', dict(authorId=authorId, per_page=per_page)))


def mirror_path(filename):
    """Determines the full path to a gutenberg text in the mirror"""
    mirror_dir = config().get_path('GUTENBERG', 'gutenberg_mirror')
    path = os.path.join(mirror_dir, filename)
    return path


def mirror_exists(file_rec):
    path = mirror_path(file_rec.file)
    return os.path.exists(path)


def textId2number(textId):
    return int(textId[5:])


def files_exist(textId, record=None):
    """Returns true if any files exist for this e-text
    in the local dataset"""
    if record is None:
        record = GutenbergBook.query.filter_by(textId=textId).first()
    if record is None:
        return False
    for x in record.gutenberg_files:
        if mirror_exists(x):
            return True
    pgid = textId2number(textId)
    if find_htmlz(pgid) is not None:
        return True
    if find_epub(pgid) is not None:
        return True
    return False


@gutenberg.route('/text/<textId>/details')
def text(textId):
    # Profiling results showing occasional lags.
    # Lags can be minimized by disabling SQLALCHEMY_ECHO and
    # and debug mode.
    # Tested no options, joinedload and subqueryload with no consistent winner.
    record = GutenbergBook.query.filter_by(textId=textId).first()
    # if blueprint has a different static_folder specified we might need to use blueprint.static_folder but currently None
    for x in record.gutenberg_files:
        if not mirror_exists(x):
            print "WARNING: Gutenberg file " + mirror_path(x.file) + " not found"
    record.gutenberg_files = filter(mirror_exists, record.gutenberg_files)

    pgid = textId2number(textId)
    if find_htmlz(pgid) is not None:
        htmlz_url = url_for('gutenberg_content_views.htmlz', pgid=pgid, path='index.html')
    else:
        htmlz_url = None
    if find_epub(pgid) is not None:
        epub_url = url_for('gutenberg_content_views.epub_ext', pgid=pgid)
    else:
        epub_url = None

    # fields format is list of tuples:
    # (Table row heading for display, gutenberg_books col name,
    #  sub-table col name if applicable)
    fields = [
        (_('Title'), 'title', ''),
        (_('Author'), 'gutenberg_creators', 'creator'),
        (_('Contributor'), 'gutenberg_contributors', 'contributor'),
        (_('Subject'), 'gutenberg_subjects', 'subject'),
        (_('Category'), 'gutenberg_categories', 'category'),
        (_('Language'), 'gutenberg_languages', 'language')
    ]
    return render_template('gutenberg/book_details.html', record=record,
                           fields=fields, epub_url=epub_url, htmlz_url=htmlz_url)


@gutenberg.route('/text/<textId>/<int:textIndex>')
def read(textId, textIndex):
    data_dir = config().get_path('GUTENBERG', 'root_dir')
    files = GutenbergFile.query.filter_by(textId=textId).all()
    assert textIndex >= 0 and textIndex < len(files)
    fullpath = safe_join(os.path.abspath(data_dir), files[textIndex].file)
    return send_file(fullpath)


def choose_file(textId):
    files = GutenbergFile.query.filter_by(textId=textId)
    #for f in files:
    #    print f
    return files[0].file


@gutenberg.route('/autocomplete')
def autocomplete():
    term = request.args.get('term', '')
    if term != '':
        index_dir = config().get_path('GUTENBERG', 'index_dir')
        ix = open_dir(index_dir)
        with ix.searcher() as searcher:
            # might use whoosh.analysis.*Analyzer to break query up
            # for matching. However it isn't clear how to combine completion
            # of partial matches across several different columns without
            # lots of effort

            # Be aware that returning a json top-level array leaves us vulnerable to CSRF.
            # http://flask.pocoo.org/docs/security/ In this case this is not of significant
            # concern because the information is not sensitive.  We use a top-level array
            # because this is what jquery autocomplete demands for use without modification.
            suggestions = get_autocomplete_matches(term)
            return Response(response=json.dumps(suggestions), mimetype="application/json")
    else:
        # Choosing an inefficient redirect because still testing different
        # approaches and its easier to centralize the handling.  If we keep
        # this approach we can just change the referencing url
        return redirect(url_for("static", filename="gutenberg_wordlist.json"))


def get_autocomplete_matches(prefix, limit=10):
    def get_prefix_like(prefix):
        (result, _) = re.subn(r'\\', u'\\\\', prefix)
        (result, _) = re.subn(r'_', u'\_', result)
        (result, _) = re.subn(r'%', u'\%', result)
        (result, _) = re.subn(r'\s+', u'%', result)
        return '%' + result + '%'

    def make_sql(colname, tablename, limit):
        sql = "SELECT {0}, downloads FROM {1} WHERE {0} LIKE :like_clause ESCAPE '\\' ORDER BY downloads LIMIT {2};".format(colname, tablename, limit)
        return sql

    like_clause = get_prefix_like(prefix)
    search_fields = [('title', 'gutenberg_books'),
                     ('creator', 'gutenberg_creators'),
                     ('contributor', 'gutenberg_contributors')]
    results = []
    with closing(db.engine.connect()) as conn:
        for colname, tablename in search_fields:
            results.extend(conn.execute(make_sql(colname, tablename, limit), like_clause=like_clause).fetchall())
    return [row[0] for row in sorted(results, key=lambda r: r[1], reverse=True)]



from flask import Blueprint, abort, send_file
import zipfile

from gutenberg_content import find_htmlz, find_epub


blueprint = Blueprint('gutenberg_content_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route('/htmlz/<int:pgid>/<path:path>')
def htmlz(pgid, path):
    htmlz_path = find_htmlz(pgid)
    if htmlz_path is None:
        print "HTMLZ Path not found " + str(pgid)
        abort(404)
    zf = zipfile.ZipFile(htmlz_path)
    f = zf.open(path)
    data = f.read()
    f.close()
    zf.close()
    return data


# This doesn't work because the relative paths are wrong
#@blueprint.route('/htmlz/<int:pgid>')
#def htmlz_index(pgid):
#    return htmlz(pgid, 'index.html')


@blueprint.route('/epub/<int:pgid>')
def epub(pgid):
    epub_path = find_epub(pgid)
    if epub_path is None:
        print "EPub Path not found " + str(pgid)
        abort(404)
    return send_file(epub_path, mimetype='application/epub+zip')


@blueprint.route('/epub/<int:pgid>.epub')
def epub_ext(pgid):
    return epub(pgid)



"""Khan Academy Video Utilities"""
import re
import os
import string


def split_path(path):
    """Parse a full file pathname into its components.
    Returns a list of (number, name) tuples.
    number may be None."""
    path = path.rstrip("\n")
    path, ext = os.path.splitext(path)
    split = [None] + re.split('(?: [-] |/)([0-9]?[0-9][0-9]) (?:- )?', path)
    assert(len(split) % 2 == 0)
    c = []
    for i in range(0, len(split), 2):
        if split[i] is not None:
            n = int(split[i])
        else:
            n = None
        c.append((n, split[i + 1]))
    return c


def assign_number_to_top_categories(paths):
    """Assign numbers to the top categories
    returned by split_path for consistency"""
    cats = {}

    def assign_number(path):
        name = path[0][1]
        n = cats.setdefault(name, len(cats) + 1)
        return [(n, name)] + path[1:]
    return map(assign_number, paths)


def categories_to_tree(categories, paths):
    """Takes output from assign_number_to_top_categories and creates a tree
    of dictionaries.  The keys are the numeric index and the values are
    tuples of (name, sub-tree)"""
    d = (None, {})
    for cat, fullpath in zip(categories, paths):
        pos = d
        for idx, (n, name) in enumerate(cat):
            if idx == len(cat) - 1:
                v = fullpath
            else:
                v = {}
            entry = pos[1].setdefault(n, (name, v))
            assert(pos[1][n][0] == name)
            pos = entry
    return d[1]


def load_paths(filename):
    """Load and parse a file of paths into a tree structure"""
    paths = [x.rstrip("\n") for x in open(filename, "r").readlines()]
    cats = [split_path(x) for x in paths]
    cats = assign_number_to_top_categories(cats)
    tree = categories_to_tree(cats, paths)
    return tree


def get(tree, s):
    """Get the tree node or full pathname for a specified
    tuple of indices such as (1, 3, 2).  Returns None if not found."""
    if len(s) == 0:
        return 'top', tree
    name, e = tree[s[0]]
    if len(s) == 1:
        return name, e
    return get(e, s[1:])


def getchildren(tree, s):
    """Returns a dictionary structure containing the parents of a
    specified node (in 'breadcrumbs'), the direct children of
    the node (in 'children') if any, and the filename if any
    (in 'file')."""
    r = {}
    name, e = get(tree, s)
    r['breadcrumbs'] = getpath(tree, s)
    if type(e) in [str, unicode]:
        r['file'] = e
    else:  # dict
        r['children'] = [(idx, subtree[0]) for (idx, subtree) in e.items()]
    return r


def getpath(tree, s):
    """Get the tree node or full pathname for a specified
    tuple of indices such as (1, 3, 2).  Returns None if not found."""
    items = []
    for i in range(1, len(s) + 1):
        name, e = get(tree, s[:i])
        items.append((s[i - 1], name))
    return items


def getfile(tree, s):
    e = getchildren(tree, s)
    return e['file']


def find(root, extension=".webm"):
    """File files with specified extension under root"""
    found = []
    cwd = os.getcwd()
    os.chdir(root)
    for path, dirs, files in os.walk('.'):
        path = path[2:]  # Remove './'
        for f in files:
            if extension is None or f[-len(extension):] == extension:
                found.append(os.path.join(unicode(path.decode('utf-8')), unicode(f.decode('utf-8'))))
    os.chdir(cwd)
    return found


def find_khan(root, extension=".webm"):
    paths = find(root, extension)
    paths.sort()  # always get a consistent category order
    cats = [split_path(x) for x in paths]
    cats = assign_number_to_top_categories(cats)
    tree = categories_to_tree(cats, paths)
    return tree


def map_symlinks(categories, paths):
    def map_path_and_cat(path, cat):
        joined = string.join([str(n) for n, name in cat], '/')
        return (path, joined)
    return map(map_path_and_cat, paths, categories)


def make_symlinks(webm_root, h264_root, symlink_root, extension=".webm"):
    """The purpose of building a tree of symlinks is to that the front end
    web server can serve the video files directly, allowing Accept-Ranges
    and other features to work with the finicky video browser clients"""
    webm_root = unicode(webm_root)
    h264_root = unicode(h264_root)
    paths = find(webm_root, extension)
    paths.sort()  # always get a consistent category order
    cats = [split_path(x) for x in paths]
    cats = assign_number_to_top_categories(cats)
    syms = map_symlinks(cats, paths)
    for src, dst in syms:
        webm_src = unicode(os.path.join(webm_root, src))
        webm_dst = unicode(os.path.join(symlink_root, dst)) + u".webm"
        directory = os.path.split(webm_dst)[0]
        if not os.path.exists(directory):
            os.makedirs(directory)
        if os.path.exists(webm_dst):
            os.remove(webm_dst)
        print webm_src.encode('utf-8'), " -> ", webm_dst.encode('utf-8')
        os.symlink(webm_src.encode('utf-8'), webm_dst.encode('utf-8'))
        h264_src = os.path.join(h264_root, os.path.splitext(src)[0] + ".m4v")
        h264_dst = os.path.join(symlink_root, dst + ".m4v")
        if os.path.exists(h264_dst):
            os.remove(h264_dst)
        print h264_src.encode('utf-8'), " -> ", h264_dst.encode('utf-8')
        assert(os.path.exists(h264_src.encode('utf-8')))
        os.symlink(h264_src.encode('utf-8'), h264_dst.encode('utf-8'))



from math import ceil

"""
Emulate flask.ext.SQLAlchemy.Pagination
Adapted from http://flask.pocoo.org/snippets/44/
"""
class Pagination(object):

    def __init__(self, page, per_page, total_count, items):
        """
        :param page: integer current page number counting from 1
        :param per_page: integer number of results per page
        :param total_count: integer number of items all together
        :param items: list of objects associated with current page of results
        """
        self.page = page
        self.per_page = per_page
        self.total_count = total_count
        self.items = items

    @property
    def pages(self):
        return int(ceil(self.total_count / float(self.per_page)))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def has_next(self):
        return self.page < self.pages

    def iter_pages(self, left_edge=2, left_current=2,
                   right_current=5, right_edge=2):
        last = 0
        for num in xrange(1, self.pages + 1):
            if num <= left_edge or \
               (num > self.page - left_current - 1 and \
                num < self.page + right_current) or \
               num > self.pages - right_edge:
                if last + 1 != num:
                    yield None
                yield num
                last = num


